# Copyright (C) 2023 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
#
"""Define pytest fixtures and such for the whole test suite"""

import pathlib
from collections import Counter
from pathlib import Path
from typing import Callable, Optional

import pytest


@pytest.fixture(scope="module")
def restore_directory(resource_path_root: Path) -> Path:
    return resource_path_root / "restores" / "tmpfpjeptku"


def assert_restore_delivered_only_expected_files(
    actual_files: list[pathlib.Path],
    expected_files: set[pathlib.Path],
    expected_dirs_to_file_counts: Counter[str],
    actual_files_to_is_file: Optional[dict[pathlib.Path, bool]] = None,
):
    """Assert that exactly the expected files were delivered

    NB: In conftest.py and not __init__.py  so that its assertions are rewritten by pytest to allow for introspection when they fail.
    See https://docs.pytest.org/en/stable/how-to/assert.html#assertion-introspection-details.
    """
    actual_is_file: Callable[[pathlib.Path], bool] = lambda f: f.is_file()
    if actual_files_to_is_file is not None:
        # actual_files' contents don't exist on disk, use given table to know which are files vs. directories
        actual_is_file = lambda f: actual_files_to_is_file[f]
    # did we get all the expected files?
    for filepath in expected_files:
        assert filepath in actual_files
    actual_dirs_to_file_counts = Counter()
    for filepath in actual_files:
        for dir, _ in expected_dirs_to_file_counts.items():
            if str(filepath).startswith(f"{dir}/") and actual_is_file(filepath):
                actual_dirs_to_file_counts[dir] += 1
    assert actual_dirs_to_file_counts == expected_dirs_to_file_counts

    # did we not get any unexpected files?
    for filepath in actual_files:
        if filepath.is_file():
            assert filepath in expected_files or any(
                str(filepath).startswith(expected_dir) for expected_dir in expected_dirs_to_file_counts.keys()
            )
