#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
""" Tests for generic and EVLA CAL manifests """

import json
import logging
import shutil
import sys
import tarfile
from pathlib import Path

# pylint: disable=C0103, C0301, E0401, E0402, R0914, R1721, W0621, W1514

from unittest.mock import patch

import pytest
from ingest_envoy.ingestion_manifest import IngestionManifestBuilder, find_manifest
from ingest_envoy.manifest_components import (
    INGESTION_ARTIFACTS_NAME,
    INIT_WEBLOG_FILENAME,
    MANIFEST_FILENAME,
    TARFILE_EXT,
    WEBLOG_FILENAME,
    AncillaryProduct,
    IngestionManifestKey,
    InputGroup,
    InputScienceProduct,
    ManifestParameters,
    OutputGroup,
    OutputScienceProduct,
    ParamsKey,
)
from ingest_envoy.utilities import AncillaryProductType, ScienceProductType, Telescope

# ingest_path is NOT unused! Don't let IJ remove the import!
from conftest import (
    EVLA_CAL_INPUT_FILENAMES,
    UNWANTED,
    find_example_manifest,
    populate_fake_evla_cal_ingest_path,
    populate_fake_final_evla_cal_ingest_path,
)

# pylint: disable=E0401, E1120


logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
logger.addHandler(logging.StreamHandler(sys.stdout))

FAKE_LOCATOR = "uid://evla/calibration/doo-wah-ditty-ditty-af123"


def test_filters_cal_input_files(ingest_path: Path):
    """
    We'll be getting science products from a directory under
    /lustre/aoc/cluster/pipeline/{CAPO_PROFILE}/workspaces/staging

    Make sure we take -only- the files to be ingested.

    :param ingest_path: our temporary dir
    """

    populate_fake_evla_cal_ingest_path(staging_dir=ingest_path)
    locator = "uid://evla/calibration/twinkle-twinkle-little-quasar"
    manifest, _ = IngestionManifestBuilder(
        telescope=Telescope.EVLA.value,
        manifest_source_dir=ingest_path,
        sp_type=ScienceProductType.EVLA_CAL.value,
        locator=locator,
    ).build()

    # does manifest contain all the data it should, and none of the data it shouldn't?

    assert manifest.locator == locator
    params = manifest.parameters

    input_group = manifest.input_group
    assert len(input_group.science_products) == 1

    output_group = manifest.output_group
    assert len(output_group.science_products) == 1
    assert len(output_group.ancillary_products) == 2
    for product in output_group.ancillary_products:
        if product.filename not in EVLA_CAL_INPUT_FILENAMES:
            assert product.filename.startswith(INGESTION_ARTIFACTS_NAME) and product.filename.endswith(TARFILE_EXT)
        assert product.filename not in UNWANTED

    sp_out = output_group.science_products[0]
    assert sp_out.type == ScienceProductType.EVLA_CAL

    assert sp_out.filename in EVLA_CAL_INPUT_FILENAMES
    assert sp_out.filename not in UNWANTED

    shutil.rmtree(ingest_path)


def test_writes_expected_evla_cal_output_files(ingest_path: Path):
    """
    Did the manifest builder produce the manifest file, the weblog, and the science product tar?

    :param ingest_path: ingestion location
    :return:
    """
    populate_fake_evla_cal_ingest_path(ingest_path)
    manifest_file, manifest = IngestionManifestBuilder(
        telescope=Telescope.EVLA.value,
        manifest_source_dir=ingest_path,
        locator="uid://evla/calibration/fee-fi-fo-fum-acdf23",
        sp_type=ScienceProductType.EVLA_CAL.value,
    ).build()
    assert manifest_file
    assert manifest

    ingestion_files = [file for file in ingest_path.iterdir()]

    # at a minimum, we expect the manifest, the ingestion artifact, and the science products tar
    assert len(ingestion_files) >= 3
    mf_json = find_manifest(ingest_path)
    assert mf_json
    tars = [file for file in ingestion_files if file.name.endswith(TARFILE_EXT)]
    assert len(tars) >= 2

    shutil.rmtree(ingest_path)


@pytest.mark.skip("TODO")
def test_writes_final_evla_cal_output(ingest_path: Path):
    """
    For a second or subsequent version of a standard calibration,
    id the manifest builder produce the manifest file, the science product tar,
    the weblog, -and- the initial weblog?

    :param ingest_path: ingestion location
    :return:
    """
    # TODO: implement this as soon as duplicate-weblog bug is fixed
    raise NotImplementedError


def test_params_json_well_formed():
    """
    Make sure our ManifestParameters makes nice JSON

    :return:
    """
    params = ManifestParameters(
        telescope=Telescope.EVLA,
        ngas_ingest=False,
        staging_source_dir=Path("/home/mchammer/evla/parallel-prod"),
    )

    params_json = params.to_dict()
    # if we can dump it, it's good
    json.dumps(params_json)


def test_input_sp_well_formed():
    """
    Make sure our InputScienceProduct makes nice JSON

    :return:
    """
    locator = "uid://evla/calibration/vanilla_heath_bar_crunch_1a23e"
    # single science product
    sp_dict = {
        "locator": locator,
    }

    sp_in = InputScienceProduct(locator=locator)
    assert sp_in.to_dict() == sp_dict


def test_input_group_well_formed():
    """
    Make sure our InputGroup makes nice JSON

    :return:
    """
    sp1 = InputScienceProduct(locator="uid://evla/execblock/coffee_heath_bar_crunch_7a23f")
    sp2 = InputScienceProduct(locator="uid://evla/execblock/mint_oreo_omg_omg")

    input_group = InputGroup(science_products=[sp1, sp2])
    assert IngestionManifestKey.SCIENCE_PRODUCTS.value in input_group.to_dict().keys()

    sps_found = input_group.science_products
    assert len(sps_found) == 2
    assert sp1 in sps_found
    assert sp2 in sps_found


def test_ancillary_product_well_formed():
    """
    The JSON shouldn't contain empty fields

    :return:
    """
    ap1 = AncillaryProduct(type=AncillaryProductType.LOG_TYPE, filename="without_feathers.tar")
    expected = {"type": AncillaryProductType.LOG_TYPE.value, "filename": ap1.filename}
    actual = ap1.to_dict()

    assert actual == expected


def test_output_group_well_formed():
    """
    Make sure our OutputScienceProduct makes nice JSON

    :return:
    """
    osp = OutputScienceProduct(type=ScienceProductType.EVLA_CAL, filename="im_a_lil_calibration.tar")
    ap1 = AncillaryProduct(type=AncillaryProductType.PIPELINE_ARTIFACTS, filename="without_feathers.tar")
    ap2 = AncillaryProduct(type=AncillaryProductType.PIPELINE_ARTIFACTS, filename="with_feathers.tar")

    opg = OutputGroup(science_products=[osp], ancillary_products=[ap1, ap2])

    # all filenames in the manifest should be bare filenames, not full paths
    for science_product in opg.science_products:
        assert "/" not in science_product.filename
    for ancillary_product in opg.ancillary_products:
        assert "/" not in ancillary_product.filename

    expected_json = {
        IngestionManifestKey.SCIENCE_PRODUCTS.value: [osp.to_dict()],
        IngestionManifestKey.ANCILLARY_PRODUCTS.value: [ap1.to_dict(), ap2.to_dict()],
    }
    actual_json = opg.to_dict()

    assert actual_json == expected_json


def test_ingestion_artifacts_tar_filename_built_just_once(ingest_path: Path):
    """
    We run into trouble if manifest builder's _build_artifacts_filename()
    is called more than once.

    :param ingest_path:
    :return:
    """
    populate_fake_evla_cal_ingest_path(ingest_path)
    with patch(
        "ingest_envoy.ingestion_manifest.IngestionManifestBuilder._build_artifacts_filename",
        return_value="ingestion_artifacts_mine_mine_mine.tar",
    ) as mock:
        # build the manifest
        IngestionManifestBuilder(
            manifest_source_dir=ingest_path,
            sp_type=ScienceProductType.EVLA_CAL.value,
            locator="uid://evla/calibration/are-we-there-yet",
            telescope=Telescope.EVLA.value,
        ).build()
        mock.assert_called_once()

    shutil.rmtree(ingest_path)


def test_ingestion_artifacts_tar_correct(ingest_path: Path):
    """
    Do we build the correct ingestion_artifacts tar for for standard imaging ingestion?

    :param ingest_path: the ingestion staging dir
    :return:
    """

    populate_fake_evla_cal_ingest_path(ingest_path)
    locator = "uid://evla/calibration/3dfa528b-9870-46c9-a200-131dbac701cc"
    # if you build it, they will come
    IngestionManifestBuilder(
        manifest_source_dir=ingest_path,
        sp_type=ScienceProductType.EVLA_CAL.value,
        locator=locator,
        telescope=Telescope.EVLA.value,
    ).build()

    # get ingestion artifacts
    artifacts_file = [file for file in ingest_path.glob("ingestion_artifacts*.tar")][0]
    with tarfile.open(artifacts_file, "r") as tar:
        # confirm that contains has as many files as we expect...
        members = tar.getmembers()

    assert len(members) == 1
    member = members[0]
    mf_path = Path(member.name)
    assert mf_path.name == MANIFEST_FILENAME

    shutil.rmtree(ingest_path)


def test_evla_cal_manifest_matches_example(ingest_path: Path):
    """
    Given the correct parameters, manifest that matches evla_cal_manifest_2021-08-02
    should be generated

    :return:
    """

    expected_dir_name = "/lustre/aoc/cluster/pipeline/dsoc-prod/stage_products/20A-346_2021_07_23_T13_37_08.376"
    example = find_example_manifest("evla_cal_manifest_2021-08-02")
    with open(example, "r") as infile:
        expected_json = dict(json.load(infile).items())

    # populate ingestion path with fake files for manifest builder to find
    populate_fake_evla_cal_ingest_path(ingest_path)

    builder = IngestionManifestBuilder(
        manifest_source_dir=ingest_path,
        telescope=Telescope.EVLA.value,
        sp_type=ScienceProductType.EVLA_CAL.value,
        locator="uid://evla/execblock/i-just-play-an-eb-on-teevee",
    )
    manifest, manifest_file = builder.build()

    with open(manifest_file, "r") as infile:
        actual_json = dict(json.load(infile).items())

    actual_json[IngestionManifestKey.PARAMETERS.value][ParamsKey.INGESTION_PATH.value] = expected_dir_name

    expected_params = expected_json["parameters"]
    actual_params = manifest.parameters.to_dict()

    assert manifest.input_group.to_dict() == expected_json["input_group"]

    expected_outgroup = expected_json["output_group"]
    expected_osp = expected_outgroup["science_products"]
    actual_osp = manifest.output_group.science_products
    assert len(actual_osp) == len(expected_osp)

    expected_aps = expected_outgroup["ancillary_products"]
    assert len(expected_aps) == 2
    actual_aps = manifest.output_group.ancillary_products
    logger.info(">>> ACTUAL ANCILLARY PRODUCTS")
    for ap in actual_aps:
        logger.info(f"{ap.filename}: {ap.type}, group with {ap.group_with}")
    assert len(actual_aps) == len(expected_aps)

    shutil.rmtree(ingest_path)


def test_evla_cal_final_manifest_finds_init_weblog(ingest_path: Path):
    """
    Given a staging dir that contains an initial -and- a final weblog,
    we expect both to be included in the ingestion manifest.

    :param ingest_path:  ingestion staging dir
    :return:
    """
    # populate ingestion path with fake files for manifest builder to find
    populate_fake_final_evla_cal_ingest_path(ingest_path)

    weblogs = []
    for file in [file for file in ingest_path.iterdir() if file.is_file()]:
        if "weblog.tgz" in file.name:
            weblogs.append(file.name)

    # be sure both files we expect are there...
    assert WEBLOG_FILENAME in weblogs and INIT_WEBLOG_FILENAME in weblogs
    # and each of them only once
    assert len(weblogs) == 2

    builder = IngestionManifestBuilder(
        manifest_source_dir=ingest_path,
        telescope=Telescope.EVLA.value,
        sp_type=ScienceProductType.EVLA_CAL.value,
        locator="uid://evla/execblock/50bb85af-ce52-49d8-b9d8-9221bfce939d",
    )
    manifest, manifest_file = builder.build()

    with open(manifest_file, "r") as infile:
        actual_json = dict(json.load(infile).items())
        og = actual_json["output_group"]
        anc_prods = og["ancillary_products"]
        assert len(anc_prods) >= 2

        weblog_filenames = []
        for ap in anc_prods:
            logger.info(f"{ap}")
            assert "filename" in ap.keys()
            if "weblog" in ap["filename"]:
                weblog_filenames.append(ap["filename"])

        assert WEBLOG_FILENAME in weblog_filenames and INIT_WEBLOG_FILENAME in weblog_filenames
        assert len(weblog_filenames) == 2


def test_evla_cal_final_manifest_matches_example(ingest_path: Path):
    """
    Given the correct parameters, manifest that matches evla_cal_final_manifest
    should be generated; i.e., ancillary)products should contain initial_weblog.tgz
    in addition to weblog.tgz

    :param ingest_path: ingestion staging dir
    :return:
    """

    expected_dir_name = "/lustre/aoc/cluster/pipeline/dsoc-prod/stage_products/20A-346_2021_07_23_T13_37_08.376"
    example = find_example_manifest("evla_cal_final_manifest")
    with open(example, "r") as infile:
        expected_json = dict(json.load(infile).items())

    # populate ingestion path with fake files for manifest builder to find
    populate_fake_final_evla_cal_ingest_path(ingest_path)

    builder = IngestionManifestBuilder(
        manifest_source_dir=ingest_path,
        telescope=Telescope.EVLA.value,
        sp_type=ScienceProductType.EVLA_CAL.value,
        locator="uid://evla/execblock/some-nonsense-not-in-db",
    )
    manifest, manifest_file = builder.build()

    with open(manifest_file, "r") as infile:
        actual_json = dict(json.load(infile).items())

    actual_json[IngestionManifestKey.PARAMETERS.value][ParamsKey.INGESTION_PATH.value] = expected_dir_name

    expected_params = expected_json["parameters"]
    actual_params = manifest.parameters.to_dict()

    assert manifest.input_group.to_dict() == expected_json["input_group"]

    expected_outgroup = expected_json["output_group"]
    expected_osp = expected_outgroup["science_products"]
    actual_osp = manifest.output_group.science_products
    assert len(actual_osp) == len(expected_osp) == 1

    for sp in actual_osp:
        logger.info(f"{sp.filename}")
        if sp.ancillary_products is not None:
            for ap in sp.ancillary_products:
                logger.info(f"{ap.filename}")

    expected_aps = expected_outgroup["ancillary_products"]
    assert len(expected_aps) == 3
    actual_aps = manifest.output_group.ancillary_products
    logger.info(">>> ancillary products")
    for ap in actual_aps:
        logger.info(f"{ap.filename}: {ap.type}, group with {ap.group_with}")
    assert len(actual_aps) == len(expected_aps)

    found_count = 0
    for ap in actual_aps:
        if ap.filename == WEBLOG_FILENAME or ap.filename == INIT_WEBLOG_FILENAME:
            found_count += 1
    assert found_count == 2

    shutil.rmtree(ingest_path)
