#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
""" Tests of CapabilityQueue """

# pylint: disable=E0110, E0401, E0611

from typing import List

import pytest

from workspaces.capability.schema import CapabilityExecution
from workspaces.capability.services.interfaces import ExecutionManagerIF, QueueReport
from workspaces.capability.services.queue_manager import CapabilityQueueManager
from workspaces.system.schema import AbstractFile

from ..conftest import mock_capability_info, mock_workflow_service


class FakeQueueReporterExecutionManager(ExecutionManagerIF):
    """
    A testing mock that pretends to be both the queue reporter and execution manager.
    Our notions of both are pretty limited, basically moving things from one list to
    another so that we can get proper reports, but otherwise not doing anything real.
    """

    submitted = []
    executing = []
    completed = []
    allowed = 3
    workflow_service = mock_workflow_service
    capability_info = mock_capability_info

    def report(self) -> List[QueueReport]:
        return [QueueReport("fake", len(self.submitted), len(self.executing), self.allowed)]

    def waiting_executions(self, queue: str, count: int) -> List[CapabilityExecution]:
        return self.submitted[:count]

    def submit_workflow_request(
        self,
        execution_id: int,
        workflow_name: str,
        workflow_args: dict,
        files: List[AbstractFile],
    ):
        execution = mock_capability_info.lookup_execution(execution_id)
        self.submitted.append(execution)

    def submit(self, execution: CapabilityExecution):
        self.submitted.append(execution)

    def execute(self, execution: CapabilityExecution):
        self.submitted.remove(execution)
        self.executing.append(execution)

    def complete(self, execution: CapabilityExecution):
        print(f">>> currently executing: {[exec.id for exec in self.executing]}")
        self.executing.remove(execution)
        self.completed.append(execution)


@pytest.mark.skip(reason="Broken by state machine rework")
def test_capability_queue_manager():
    # let's start by putting five things into the execution manager
    test_reporter_executor = FakeQueueReporterExecutionManager()
    executions = []
    for i in range(1, 6):
        execution = CapabilityExecution(id=i)
        test_reporter_executor.submit(execution)
        executions.append(execution)

    for i in range(1, 3):
        execution = test_reporter_executor.submitted[i]
        test_reporter_executor.execute(execution)
        executions.append(execution)

    # now let's make our queue manager and see what the report tells us
    queue = CapabilityQueueManager(test_reporter_executor, test_reporter_executor)
    queue.launch_if_availability()

    # we should now have 3 executing and 2 submitted
    assert test_reporter_executor.report()[0].executing == 3
    assert test_reporter_executor.report()[0].waiting == 2

    # now let's finish a couple of them
    test_reporter_executor.complete(executions[1])
    queue.execution_complete()
    test_reporter_executor.complete(executions[3])
    queue.execution_complete()

    # we should now have 3 executing and 0 submitted
    assert test_reporter_executor.report()[0].executing == 3
    assert test_reporter_executor.report()[0].waiting == 0

    # let's complete a couple more, this time no notification in between
    test_reporter_executor.complete(executions[2])
    test_reporter_executor.complete(executions[4])
    queue.execution_complete()

    # we should now have 1 executing and 0 submitted
    assert test_reporter_executor.report()[0].executing == 1
    assert test_reporter_executor.report()[0].waiting == 0

    # let's add four more
    for i in range(6, 10):
        execution = CapabilityExecution(id=i)
        test_reporter_executor.submit(execution)
        executions.append(execution)

    queue.execution_requested()

    # now check again and we should have 3 executing and 2 submitted again
    assert test_reporter_executor.report()[0].executing == 3
    assert test_reporter_executor.report()[0].waiting == 2
